#include "ImageRgb.h"
#include "XFile.h"

namespace Torch {

ImageRgb::ImageRgb()
{
	n_planes = 3;
	coding = "rgb";
}

ImageRgb::ImageRgb(const char *filename_)
{
	n_planes = 3;
	coding = "rgb";

	load(filename_);
}

ImageRgb::ImageRgb(int width_, int height_) : Image(width_, height_, 3)
{
	coding = "rgb";
}

bool ImageRgb::copyFrom(Image *image_)
{
	return copyFrom(image_->width, image_->height, image_->pixmap, image_->coding);
}

bool ImageRgb::copyFrom(int width_, int height_, unsigned char *pixmap_, const char *coding_)
{
   	width = width_;
	height = height_;
	image_size = width * height;
	frame_size = image_size;

	imageRealloc();
	
	if(strcmp(coding_, "gray") == 0)
	{
	   	if(verbose) message("ImageRgb::copyFrom(uchar *) convert from gray");
		
	   	for(int i = 0 ; i < image_size ; i++)
		{
		   	unsigned char pixel = pixmap_[i];
			int offset = 3 * i;
			
			pixmap[offset] = pixel;
			pixmap[offset+1] = pixel;
			pixmap[offset+2] = pixel;
		}
	   
		updateDataFromPixmap();

		return true;
	}
	else if(strcmp(coding_, "rgb") == 0)
	{
	   	if(verbose) message("ImageRgb::copyFrom(uchar *) copy from rgb");
		
	   	for(int i = 0 ; i < 3 * image_size ; i++) pixmap[i] = pixmap_[i];
		
		updateDataFromPixmap();

		return true;
	}
	else if(strcmp(coding_, "yuv") == 0)
	{
	   	if(verbose) message("ImageRgb::copyFrom(uchar *) convert from yuv (not implemented yet !)");
		
		updateDataFromPixmap();

		return true;
	}

   	return false;
}

bool ImageRgb::copyFrom(int width_, int height_, real *data_, const char *coding_, real max_)
{
   	width = width_;
	height = height_;
	image_size = width * height;
	frame_size = image_size;

	imageRealloc();
	
	real rescale = 255.0 / max_;
	
	if(strcmp(coding_, "gray") == 0)
	{
	   	if(verbose) message("ImageRgb::copyFrom(real *) convert from gray");
		
	   	for(int i = 0 ; i < image_size ; i++)
		{
		   	real pixel = data_[i] * rescale;
			int offset = 3 * i;
			
			data[offset] = pixel;
			data[offset+1] = pixel;
			data[offset+2] = pixel;
		}

		updatePixmapFromData();

		return true;
	}
	else if(strcmp(coding_, "rgb") == 0)
	{
	   	if(verbose) message("ImageRgb::copyFrom(real *) copy from rgb");
		
	   	for(int i = 0 ; i < 3 * image_size ; i++) data[i] = data_[i] * rescale;

		updatePixmapFromData();

		return true;
	}
	else if(strcmp(coding_, "yuv") == 0)
	{
	   	if(verbose) message("ImageRgb::copyFrom(real *) convert from yuv (not implemented yet !)");
		
		updatePixmapFromData();

		return true;
	}

   	return false;
}

void ImageRgb::saveXFile(XFile *file)
{
	// Save as ppm (P6 format)

	file->printf("P6\n");
	file->printf("#\n");
	file->printf("# Image generated using Torch vision\n");
	file->printf("# (c) 2004-2005 Sebastien Marcel marcel@idiap.ch\n");
	file->printf("# IDIAP Research Institute\n");
	file->printf("#\n");
	file->printf("%d %d\n", width, height);
	file->printf("255\n");
   
	file->write(pixmap, sizeof(unsigned char), n_planes*width*height);
}

void ImageRgb::loadXFile(XFile *file)
{
	// Load as ppm (P6 format)
	
	char buffer[200];

   	// Reads Magic Number
	if(!file->gets(buffer, 200)) error("ImageRgb::loadXFile() not enable to read ppm file\n");
	if(strcmp(buffer, "P6\n") != 0) error("Incorrect magic number\n");

	// Reads comments
	if(!file->gets(buffer, 200)) error("ImageRgb::loadXFile() not enable to read ppm file\n");
	//while(buffer[0] == '#') if(!file->gets(buffer, 200)) error("ImageRgb::loadXFile() not enable to read ppm file\n");
	while(buffer[0] == '#' || buffer[0] == '\n') if(!file->gets(buffer, 200)) error("ImageRgb::loadXFile() not enable to read ppm file\n");

	//
	// Reads header
   	sscanf(buffer, "%d %d\n", &width, &height);
   	if(width <= 0) error("ImageRgb::loadXFile() incorrect width in ppm file\n");
   	if(height <= 0) error("ImageRgb::loadXFile() incorrect height in ppm file\n");

	image_size = width * height;
	frame_size = image_size;

	//
	// Reads resolution
	if(!file->gets(buffer, 200)) error("ImageRgb::loadXFile() not enable to read ppm file\n");
	if(strcmp(buffer, "255\n") != 0) error("ImageRgb::loadXFile() incorrect resolution, only 255 is supported\n");

	//
	// Allocate/Reallocate memory if needed
	imageRealloc();
	
	//
	// Reads the pixmap
	int n_ = file->read(pixmap, sizeof(unsigned char), 3 * image_size);
	if(n_ != 3 * image_size) error("ImageRgb::loadXFile() impossible to read the image (%d bytes read != %d bytes expected)\n", n_, 3 * image_size);

	//
	// Update image data
	updateDataFromPixmap();
}

void ImageRgb::drawpixel(int x, int y, Color c)
{
	if(x>=0 && y>=0 && x<width && y<height)
	{
		/// Warning the color should be appropriate

		int offset_0 = 3 *(y * width + x);
		int offset_1 = offset_0+1;
		int offset_2 = offset_0+2;

		pixmap[offset_0] = c.data0;
		pixmap[offset_1] = c.data1;
		pixmap[offset_2] = c.data2;

		data[offset_0] = c.data0;
		data[offset_1] = c.data1;
		data[offset_2] = c.data2;
	}
}

ImageRgb::~ImageRgb()
{
}

}

