#ifndef IP_ONEBLOB_INC
#define IP_ONEBLOB_INC

#include "ipColorCube.h"

namespace Torch {

/** This class is designed to extract Blob features assuming 1 blob in the image

    An image is filtered using a colorcube and the parameters of the blob are computed.

    \begin{itemize}
    \item the center of the blob ($\mu_x$ and $\mu_y$),
    \item the excentricity of the blob ($e = \sqrt{1-\frac{b^2}{a^2}}$),
    \item the angle of the major axis of the blob with the horizontal,
    \item the surface of the blob ($s = \pi \times ab$).
    \end{itemize}
    
    \begin{verbatim}

      seq_in->frames[0]        +--------------+
	       +----+          |              |
	       |    |   ---->  |  ipOneBlob   | ----> seq_out->frames[0][0] is the x mean
	       |    |          |              |       seq_out->frames[0][1] is the y mean
	       +----+          +--------------+       seq_out->frames[0][2] is the excentricity
                image                                 seq_out->frames[0][3] is the angle
	                                              seq_out->frames[0][4] is the surface
						      
    \end{verbatim}
    
    @author Agnes Just (just@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
    
    @see ipColorCube
*/
class ipOneBlob : public ipCore
{
   	// width of the input image
	int width;

	// height of the input image
	int height;
	
	// the ipCore to filter the image 
	ipColorCube *ipcolorcube;
    
public:
	/** constructor
	    
	    @param width_ is the #width# of the input image
	    @param height_ is the #height# of the input image
	    @param colorcube_filename is the filename of the colorcube to use
	    @param coding_ is the #coding# format of the input image
	*/
	ipOneBlob(int width_, int height_, const char *colorcube_filename, const char *coding_);
	
	///
	virtual ~ipOneBlob();

	///
	virtual void process(Sequence *seq_in);
};
  
}

#endif
