#include "ipSWhaar.h"
#include "HaarStumpMachine.h"
#include "HaarRealStumpMachine.h"
#include "DiskXFile.h"


namespace Torch {

ipSWhaar::ipSWhaar(	
		real *integralimage_,
		real *integralimage2_,
		int width_image_, int height_image_, 
		int width_pattern_, int height_pattern_,
		const char *model_filename_, real threshold_,
		bool verbose,
		bool realstump) 
		: ipSubWindow(width_image_, height_image_, width_pattern_, height_pattern_, threshold_)
{
	ii = integralimage_;
	ii2 = integralimage2_;
	
	seq_input = new(allocator) Sequence(&ii, 1, width_image * height_image);

	/*************   load the model **************/
	DiskXFile *file;
        file = new(allocator) DiskXFile(model_filename_, "r");
	
	int width_, height_;
	file->taggedRead(&width_, sizeof(int), 1, "WIDTH");
        file->taggedRead(&height_, sizeof(int), 1, "HEIGHT");
	if(width_ != width_pattern || height_ != height_pattern)
		error("ipSWhaar: model wrong size");
	
        file->taggedRead(&n_stages, sizeof(int), 1, "N_STAGES");

	machines = (Machine ***)allocator->alloc(n_stages*sizeof(Machine **));
	
	thresholds = (real *)allocator->alloc(n_stages*sizeof(real));
	n_trainers = (int *)allocator->alloc(n_stages*sizeof(int));
	weights = (real **)allocator->alloc(n_stages*sizeof(real *));
	
	for(int i = 0; i < n_stages; i++)
	{
		file->taggedRead(&thresholds[i], sizeof(real), 1, "THRESHOLD");
                file->taggedRead(&n_trainers[i], sizeof(int), 1, "N_TRAINERS");	

		weights[i] = (real *)allocator->alloc(n_trainers[i]*sizeof(real));
		
		machines[i] = (Machine **)allocator->alloc(n_trainers[i]*sizeof(Machine *));
		for(int j = 0; j < n_trainers[i]; j++)
		   	if(realstump)
				machines[i][j] = new(allocator) HaarRealStumpMachine(width_, height_, ii2);
			else machines[i][j] = new(allocator) HaarStumpMachine(width_, height_, ii2);
	}

	// i.e. ImageWeightedSumMachine->loadXFile()
	for(int i = 0; i < n_stages; i++)
	{
		file->taggedRead(weights[i], sizeof(real), n_trainers[i], "WEIGHTS");

		for (int j = 0; j < n_trainers[i]; j++)
			machines[i][j]->loadXFile(file);
	}		

	
	/*********************************************/
	
	n_rejections_stage = (int *)allocator->alloc(n_stages*sizeof(int));
	for(int i=0; i<n_stages; i++)
		n_rejections_stage[i] = 0;
	
	if(verbose)
	{		
		print("ipSWhaar::ipSWhaar()\n");
		print("   width_image=%d\n", width_image);
		print("   height_image=%d\n", height_image);
		print("   width_pattern=%d\n", width_pattern);
		print("   height_pattern=%d\n", height_pattern);
		print("\n");

		print("Loading model %s ...\n", model_filename_);
		print("Cascade (%d stages):\n", n_stages);
		for(int i=0; i<n_stages; i++)
		{
		    print(" + stage %d:\n", i);
		    print("\t + n_classifiers: %d\n", n_trainers[i]);
		    print("\t + threshold: %g\n", thresholds[i]);
		}
	}


	// init masks[][]
	masks = (ipHaar ***)allocator->alloc(n_stages*sizeof(ipHaar **));
	for(int i=0; i<n_stages; i++)
	{
		masks[i] = (ipHaar **)allocator->alloc(n_trainers[i]*sizeof(ipHaar *));
		for(int j=0; j<n_trainers[i]; j++)
		{
		   	if(realstump)
			{
				HaarRealStumpMachine *m_ = (HaarRealStumpMachine *) machines[i][j];
				masks[i][j] = m_->the_mask;
			}
			else
			{
				HaarStumpMachine *m_ = (HaarStumpMachine *) machines[i][j];
				masks[i][j] = m_->the_mask;
			}
		}
	}

	// resetImage
	for(int i=0; i<n_stages; i++)
	    for(int j=0; j<n_trainers[i]; j++)
		masks[i][j]->resetImage(width_image, height_image);
}

void ipSWhaar::process(Sequence *seq_in)
{
   	ispattern = true;
	
	for(int i=0; i<n_stages; i++)
	{
	    real output_iwsm = 0.0;

	    for(int j=0; j<n_trainers[i]; j++)
	    {
		// set window size and x,y
		masks[i][j]->resetWindowSize(w_window, h_window);
		masks[i][j]->resetWindowXY(x_window, y_window);
	    	
		//forward
		machines[i][j]->forward(seq_input);

		//get the output
		real output_ = machines[i][j]->outputs->frames[0][0];

		output_iwsm += weights[i][j] * output_;
	    }
		
	    if(output_iwsm < thresholds[i])
	    {
	       	    ispattern = false;
		    seq_out->frames[0][0] = -INF;
		    n_rejections_stage[i]++;
		    break;
	    }
	    else
		    seq_out->frames[0][0] = output_iwsm;
	}
	
	if(ispattern && seq_out->frames[0][0] < threshold)
	{
		ispattern = false;
		seq_out->frames[0][0] = -INF;
		n_rejections_stage[n_stages-1]++;
	}
}

void ipSWhaar::stat()
{
	print("    rejections per stages:\n");
	int total_ = 0;
	for(int i=0; i<n_stages; i++) 
		total_ += n_rejections_stage[i];
	for(int i=0; i<n_stages; i++)
	{
		real z = 100.0 * (real) n_rejections_stage[i] / (real) total_;
		print("      + stage %d: %d \t %2.2f %\n", i, n_rejections_stage[i], z);
	}
	print("\n");
	//print("    total number of rejections : %d\n", total_);
}

ipSWhaar::~ipSWhaar()
{
}

}

