#include "ipSWmlpCascade.h"
#include "DiskXFile.h"

namespace Torch {

ipSWmlpCascade::ipSWmlpCascade(int width_image_, int height_image_, 
		 int width_pattern_, int height_pattern_,
		 const char *model_filename_, real threshold_,
		 bool verbose_) 
		: ipSubWindow(width_image_, height_image_, width_pattern_, height_pattern_, threshold_)
{
   	verbose = verbose_;
	
	if(verbose)
	{
		print("ipSWmlpCascade::ipSWmlpCascade()\n");
		print("   width_image=%d\n", width_image);
		print("   height_image=%d\n", height_image);
		print("   width_pattern=%d\n", width_pattern);
		print("   height_pattern=%d\n", height_pattern);
	}

	size_pattern = width_pattern * height_pattern;
 
	sub_image = new(allocator) Sequence(1, size_pattern);	

	grayimage = new(allocator) ImageGray(width_pattern, height_pattern);

	
	/*************   load the model **************/
	DiskXFile *model = new DiskXFile(model_filename_, "r");
	
	int w_, h_;

	model->taggedRead(&w_, sizeof(int), 1, "WIDTH");
	model->taggedRead(&h_, sizeof(int), 1, "HEIGHT");
	model->taggedRead(&n_stages, sizeof(int), 1, "N_STAGES");

	if((w_ != width_pattern) || (h_ != height_pattern))
		error("ipSWmlpCascade() incorrect width (%d == %d ?) or height (%d == %d ?)", w_, width_pattern, h_, height_pattern);
	

	if(verbose)
	{
		print(" width = %d\n", width_pattern);
		print(" height = %d\n", height_pattern);
		print(" n_stages = %d\n", n_stages);
	}
	
	mv_norm = new(allocator) MyMeanVarNorm(size_pattern, 1);
  
	mlps = (Machine **)allocator->alloc(sizeof(Machine *)*n_stages);
	thresholds = (real *)allocator->alloc(sizeof(real)*n_stages);

	int n_params = 0;
	for(int s = 0 ; s < n_stages ; s++)
	{
	   	//
		model->taggedRead(&thresholds[s], sizeof(real), 1, "THRESHOLD");
		if(verbose) print(" loading MLP %d (%2.9f)\n", s, thresholds[s]);
		
		//
		MyMLP *mlp_ = new(allocator) MyMLP();
		mlp_->verbose = verbose;
		mlp_->loadXFile(model);
		mlps[s] = mlp_;
		n_params += mlp_->params->n_params;

		if(verbose) mlp_->info();
	}
	mv_norm->loadXFile(model);
	delete model;

	//
	if(verbose) message("Number of parameters of the cascade: %d", n_params);
	
	//
	enhancing = new(allocator) ipHistoEqual(width_pattern, height_pattern, "float");
	smoothing = new(allocator) ipSmoothGaussian3(width_pattern, height_pattern, "gray", 0.25);
}

void ipSWmlpCascade::process(Sequence *seq_in)
{
   	ispattern = true;

	cropGray(seq_in->frames[0], width_image, height_image, x_window, y_window, w_window, h_window, sub_image->frames[0]);

        enhancing->process(sub_image);
        smoothing->process(enhancing->seq_out);
	
	mv_norm->preProcessInputs(smoothing->seq_out);
	
	for(int i=0; i<n_stages; i++)
	{
		mlps[i]->forward(smoothing->seq_out);

		//get the output
		real output_ = mlps[i]->outputs->frames[0][0];
	    
		if(output_ < thresholds[i])
		{
		   	ispattern = false;
			seq_out->frames[0][0] = -INF;
			break;
	    	}
	    	else seq_out->frames[0][0] = output_;
	}
	   
	if(ispattern && seq_out->frames[0][0] >= threshold)
	{
		if(verbose)
			print(" -> mlp = %g\n", seq_out->frames[0][0]);

		if(verbose)
		{
			char str[255];

			sprintf(str, "subimage-%d-%d-%d-%d-%g.pgm", x_window, y_window, w_window, h_window, seq_out->frames[0][0]);

			grayimage->copyFrom(w_window, h_window, smoothing->seq_out->frames[0], "gray", 1.0);
			grayimage->save(str);
		}
	}
	else
	{
	   	ispattern = false;
		seq_out->frames[0][0] = -INF;
	}
}

ipSWmlpCascade::~ipSWmlpCascade()
{
}

}

